@@ -19,9 +19,8 @@ package dns_handler
 
 import (
 	"context"
-	"github.com/IrineSistiana/mosdns/dispatcher/handler"
-	"github.com/IrineSistiana/mosdns/dispatcher/pkg/concurrent_limiter"
-	"github.com/IrineSistiana/mosdns/dispatcher/pkg/executable_seq"
+	"github.com/IrineSistiana/mosdns/v2/dispatcher/handler"
+	"github.com/IrineSistiana/mosdns/v2/dispatcher/pkg/concurrent_limiter"
 	"github.com/miekg/dns"
 	"go.uber.org/zap"
 	"sync"
@@ -43,12 +42,12 @@ type DefaultHandler struct {
 	Logger *zap.Logger
 
 	// Entry is the entry ExecutablePlugin's tag. This cannot be nil.
-	Entry executable_seq.ExecutableCmd
+	Entry handler.ExecutableChainNode
 
 	// ConcurrentLimit controls the max concurrent queries for the DefaultHandler.
 	// If ConcurrentLimit <= 0, means no limit.
 	// When calling DefaultHandler.ServeDNS(), if a query exceeds the limit, it will wait on a FIFO queue until
-	// - its ctx is done -> The query will be dropped silently.
+	// - its ctx is done or currently there are more than 3 x ConcurrentLimit queries waiting -> The query will be dropped silently.
 	// - it can be proceeded -> Normal procedure.
 	ConcurrentLimit int
 
@@ -58,7 +57,7 @@ type DefaultHandler struct {
 }
 
 // ServeDNS
-// If entry returns an err, a SERVFAIL response will be sent back to client.
+// If entry returns an error, a SERVFAIL response will be sent back to client.
 // If concurrentLimit is reached, the query will block and wait available token until ctx is done.
 func (h *DefaultHandler) ServeDNS(ctx context.Context, qCtx *handler.Context, w ResponseWriter) {
 	h.initOnce.Do(func() {
@@ -73,17 +72,19 @@ func (h *DefaultHandler) ServeDNS(ctx context.Context, qCtx *handler.Context, w
 	})
 
 	if h.limiter != nil {
+		if int(h.limiter.Running()) > h.limiter.Max() { // too many waiting query, silently drop it.
+			return
+		}
+
 		select {
 		case h.limiter.Wait() <- struct{}{}:
 			defer h.limiter.Done()
 		case <-ctx.Done():
-			// silently drop this query
-			return
+			return // ctx timeout, silently drop it.
 		}
 	}
 
-	err := h.execEntry(ctx, qCtx)
-
+	err := handler.ExecChainNode(ctx, qCtx, h.Entry)
 	if err != nil {
 		h.logger.Warn("entry returned an err", qCtx.InfoField(), zap.Error(err))
 	} else {
@@ -106,10 +107,6 @@ func (h *DefaultHandler) ServeDNS(ctx context.Context, qCtx *handler.Context, w
 	}
 }
 
-func (h *DefaultHandler) execEntry(ctx context.Context, qCtx *handler.Context) error {
-	return executable_seq.ExecRoot(ctx, qCtx, h.logger, h.Entry)
-}
-
 type DummyServerHandler struct {
 	T       *testing.T
 	WantMsg *dns.Msg
