@@ -21,10 +21,10 @@ import (
 	"crypto/rand"
 	"crypto/tls"
 	"github.com/AdguardTeam/dnsproxy/upstream"
-	"github.com/IrineSistiana/mosdns/dispatcher/pkg/dnsutils"
-	"github.com/IrineSistiana/mosdns/dispatcher/pkg/server/dns_handler"
-	"github.com/IrineSistiana/mosdns/dispatcher/pkg/server/http_handler"
-	"github.com/IrineSistiana/mosdns/dispatcher/pkg/utils"
+	"github.com/IrineSistiana/mosdns/v2/dispatcher/pkg/dnsutils"
+	"github.com/IrineSistiana/mosdns/v2/dispatcher/pkg/server/dns_handler"
+	"github.com/IrineSistiana/mosdns/v2/dispatcher/pkg/server/http_handler"
+	"github.com/IrineSistiana/mosdns/v2/dispatcher/pkg/utils"
 	"github.com/miekg/dns"
 	"net"
 	"sync"
