@@ -22,9 +22,9 @@ import (
 	"crypto/x509"
 	"errors"
 	"fmt"
-	"github.com/IrineSistiana/mosdns/dispatcher/handler"
-	"github.com/IrineSistiana/mosdns/dispatcher/pkg/upstream"
-	"github.com/IrineSistiana/mosdns/dispatcher/pkg/utils"
+	"github.com/IrineSistiana/mosdns/v2/dispatcher/handler"
+	"github.com/IrineSistiana/mosdns/v2/dispatcher/pkg/upstream"
+	"github.com/IrineSistiana/mosdns/v2/dispatcher/pkg/utils"
 	"github.com/miekg/dns"
 	"time"
 )
@@ -42,15 +42,12 @@ type fastForward struct {
 	args *Args
 
 	upstream []utils.Upstream
-
-	sfGroup utils.ExchangeSingleFlightGroup
 }
 
 type Args struct {
-	Deduplicate bool              `yaml:"deduplicate"`
-	CA          []string          `yaml:"ca"`
-	Timeout     int               `yaml:"timeout"`
-	Upstream    []*UpstreamConfig `yaml:"upstream"`
+	Upstream []*UpstreamConfig `yaml:"upstream"`
+	Timeout  int               `yaml:"timeout"`
+	CA       []string          `yaml:"ca"`
 }
 
 type UpstreamConfig struct {
@@ -62,11 +59,6 @@ type UpstreamConfig struct {
 	IdleTimeout        int  `yaml:"idle_timeout"`
 	MaxConns           int  `yaml:"max_conns"`
 	InsecureSkipVerify bool `yaml:"insecure_skip_verify"`
-
-	Protocol   string `yaml:"protocol"`    // Deprecated
-	ServerName string `yaml:"server_name"` // Deprecated
-	URL        string `yaml:"url"`         // Deprecated
-	Timeout    int    `yaml:"timeout"`     // Deprecated
 }
 
 func Init(bp *handler.BP, args interface{}) (p handler.Plugin, err error) {
@@ -99,56 +91,17 @@ func newFastForward(bp *handler.BP, args *Args) (*fastForward, error) {
 			return nil, errors.New("missing server addr")
 		}
 
-		var u *upstream.FastUpstream
-		var err error
-
-		// TODO: Remove this after v2.0
-		// compatible mode
-		if len(config.Protocol) != 0 {
-			var addr string
-			var dialAddr string
-			switch config.Protocol {
-			case "udp", "tcp":
-				addr = config.Protocol + "://" + config.Addr
-			case "dot":
-				if len(config.ServerName) == 0 {
-					return nil, errors.New("missing server name")
-				}
-				addr = "tls://" + config.ServerName
-				dialAddr = config.Addr
-			case "doh":
-				if len(config.URL) == 0 {
-					return nil, errors.New("missing server url")
-				}
-				addr = config.URL
-				dialAddr = config.Addr
-			default:
-				return nil, fmt.Errorf("invalid protocol %s", config.Protocol)
-			}
-			u, err = upstream.NewFastUpstream(
-				addr,
-				upstream.WithDialAddr(dialAddr),
-				upstream.WithSocks5(config.Socks5),
-				upstream.WithReadTimeout(time.Duration(args.Timeout)*time.Second),
-				upstream.WithIdleTimeout(time.Duration(config.IdleTimeout)*time.Second),
-				upstream.WithMaxConns(config.MaxConns),
-				upstream.WithInsecureSkipVerify(config.InsecureSkipVerify),
-				upstream.WithRootCAs(rootCAs),
-				upstream.WithLogger(bp.L()),
-			)
-		} else {
-			u, err = upstream.NewFastUpstream(
-				config.Addr,
-				upstream.WithDialAddr(config.DialAddr),
-				upstream.WithSocks5(config.Socks5),
-				upstream.WithReadTimeout(time.Duration(args.Timeout)*time.Second),
-				upstream.WithIdleTimeout(time.Duration(config.IdleTimeout)*time.Second),
-				upstream.WithMaxConns(config.MaxConns),
-				upstream.WithInsecureSkipVerify(config.InsecureSkipVerify),
-				upstream.WithRootCAs(rootCAs),
-				upstream.WithLogger(bp.L()),
-			)
-		}
+		u, err := upstream.NewFastUpstream(
+			config.Addr,
+			upstream.WithDialAddr(config.DialAddr),
+			upstream.WithSocks5(config.Socks5),
+			upstream.WithReadTimeout(time.Duration(args.Timeout)*time.Second),
+			upstream.WithIdleTimeout(time.Duration(config.IdleTimeout)*time.Second),
+			upstream.WithMaxConns(config.MaxConns),
+			upstream.WithInsecureSkipVerify(config.InsecureSkipVerify),
+			upstream.WithRootCAs(rootCAs),
+			upstream.WithLogger(bp.L()),
+		)
 
 		if err != nil {
 			return nil, fmt.Errorf("failed to init upstream: %w", err)
@@ -186,12 +139,17 @@ func (u *upstreamWrapper) Trusted() bool {
 // qCtx.Status() will be set as
 // - handler.ContextStatusResponded: if it received a response.
 // - handler.ContextStatusServerFailed: if all upstreams failed.
-func (f *fastForward) Exec(ctx context.Context, qCtx *handler.Context) (err error) {
-	return f.exec(ctx, qCtx)
+func (f *fastForward) Exec(ctx context.Context, qCtx *handler.Context, next handler.ExecutableChainNode) error {
+	err := f.exec(ctx, qCtx)
+	if err != nil {
+		return err
+	}
+
+	return handler.ExecChainNode(ctx, qCtx, next)
 }
 
 func (f *fastForward) exec(ctx context.Context, qCtx *handler.Context) (err error) {
-	r, err := f.exchange(ctx, qCtx)
+	r, err := utils.ExchangeParallel(ctx, qCtx, f.upstream, f.L())
 	if err != nil {
 		qCtx.SetResponse(nil, handler.ContextStatusServerFailed)
 		return err
@@ -200,10 +158,3 @@ func (f *fastForward) exec(ctx context.Context, qCtx *handler.Context) (err erro
 	qCtx.SetResponse(r, handler.ContextStatusResponded)
 	return nil
 }
-
-func (f *fastForward) exchange(ctx context.Context, qCtx *handler.Context) (r *dns.Msg, err error) {
-	if f.args.Deduplicate {
-		return f.sfGroup.Exchange(ctx, qCtx, f.upstream, f.L())
-	}
-	return utils.ExchangeParallel(ctx, qCtx, f.upstream, f.L())
-}
